@@ -1,11 +1,11 @@
 import io
 from abc import ABCMeta
 from enum import IntEnum
-from typing import BinaryIO, List, Callable, Optional, TypeVar, Type, Dict, Union
+from typing import BinaryIO, List, Callable, Optional, TypeVar, Type, Dict, Union, Set
 
 from protox.encoding import decode_header, wire_type_to_decoder
 from protox.exceptions import MessageEncodeError, MessageDecodeError, FieldValidationError
-from protox.fields import Field, OneOf, Repeated, MessageField, MapField
+from protox.fields import Field, OneOf, Repeated, MessageField, MapField, PrimitiveField
 from protox.validated_dict import ValidatedDict
 from protox.validated_list import ValidatedList
 
@@ -19,7 +19,7 @@ __all__ = [
 
 
 class FieldGetter:
-    def __init__(self, key, default):
+    def __init__(self, key: str, default=None):
         self._key = key
         self._default = default
 
@@ -104,11 +104,16 @@ def _add_field_to_message(
         field_getter = RepeatedGetter(name, field.field)
     elif isinstance(field, MapField):
         field_getter = MapGetter(name, field)
-    else:
+    elif isinstance(field, MessageField):
+        # TODO: probably replace with MessageGetter
+        field_getter = FieldGetter(name, None)
+    elif isinstance(field, PrimitiveField):
         field_getter = FieldGetter(name, field.default)
+    else:
+        raise Exception('Unreachable code')
 
     if name in message_type._one_of_by_field_name:
-        if field.required:
+        if getattr(field, 'required', False):
             raise FieldValidationError(
                 f'A one_of field {message_type.__name__}.{name} should be optional, not required!'
             )
@@ -126,6 +131,12 @@ def _add_field_to_message(
 
     field.name = name
 
+    if (
+        getattr(field, 'required', False)
+        and getattr(field, 'default', None) is None
+    ):
+        message_type._required_fields.add(field.name)
+
     setattr(
         message_type,
         name,
@@ -163,6 +174,7 @@ class MessageMeta(ABCMeta):
 
             new_cls._field_by_name = {}
             new_cls._field_by_number = {}
+            new_cls._required_fields = set()
 
             one_ofs = mcs._collect_one_ofs(name, namespace)
             new_cls._one_of_by_field_name = one_ofs
@@ -206,7 +218,7 @@ class MessageMeta(ABCMeta):
 
                 field_numbers.add(value.number)
 
-                if value.default is not None:
+                if hasattr(value, 'default') and value.default is not None:
                     try:
                         value.validate_value(value.default)
                     except ValueError:
@@ -227,6 +239,7 @@ class Message(metaclass=MessageMeta):
     _field_by_name: Dict[str, Union[Field, Repeated]] = None
     _field_by_number: Dict[int, Union[Field, Repeated]] = None
     _one_of_by_field_name: dict = None
+    _required_fields: Set[str] = None
 
     # Provided by code generator
     DESCRIPTOR: 'DescriptorProto' = None
@@ -280,7 +293,7 @@ class Message(metaclass=MessageMeta):
         return MessageField(
             cls,
             number=number,
-            required=required
+            required=required,
         )
 
     @classmethod
@@ -326,8 +339,10 @@ class Message(metaclass=MessageMeta):
                 # read and discard unknown field
                 wire_type_to_decoder[wire_type](stream)
 
+        # TODO: when adding field to Message if the field is required
+        #  put it to Message._required_fields to simplify the following check
         for key, field, in cls._field_by_name.items():
-            if field.required and not field.default and key not in message_fields:
+            if getattr(field, 'required', False) and not getattr(field, 'default', None) and key not in message_fields:
                 raise MessageDecodeError(
                     f"Field {cls.__name__}.{key} is required but was not read from input stream"
                 )
@@ -341,16 +356,18 @@ class Message(metaclass=MessageMeta):
         return stream.read()
 
     def to_stream(self, stream: BinaryIO):
+        if not self.is_initialized():
+            required_fields = self._required_fields - set(self._field_by_name.keys())
+
+            raise MessageEncodeError(
+                f'Fields {required_fields} required but was not set'
+            )
+
         for key, field in self._field_by_name.items():
             value = getattr(self, key)
 
-            if value is None:
-                if field.required:
-                    raise MessageEncodeError(
-                        f'Field {type(self).__name__}.{key} is required but was not set'
-                    )
-                else:
-                    continue
+            if value is None and not getattr(field, 'required', False):
+                continue
 
             stream.write(field.encode(value))
 
@@ -387,13 +404,11 @@ class Message(metaclass=MessageMeta):
 
     def is_initialized(self) -> bool:
         """
-        Returns True if all required fields are set
+        Returns True if all required non-default fields are set
         """
-        for name, field in self._field_by_name.items():
-            if field.required and name not in self._data:
-                return False
-
-        return True
+        return self._required_fields.issubset(
+            self._data.keys()
+        )
 
     @classmethod
     def list_fields(cls) -> List[str]:
