@@ -1,10 +1,9 @@
-import json
 from enum import IntEnum
-from typing import Dict, List, Iterable, Union, MutableMapping, Iterator
-
-from protox.message import MessageMeta
+from typing import Dict, List, Iterable, Union, Iterator, BinaryIO, MutableMapping
 
 from protox import Message, fields, one_of, define_fields
+from protox.encoding import decode_header
+from protox.exceptions import MessageDecodeError
 
 PyValue_T = Union[
     type(None),
@@ -19,20 +18,15 @@ PyValue_T = Union[
 
 class Struct(Message, MutableMapping):
     _fields: Dict[str, 'Value']
-    __metaclass__ = MessageMeta
 
-    def __init__(self, initial_value: Dict[str, PyValue_T] = None, **kwargs):
+    def __init__(self, **kwargs: PyValue_T):
         super().__init__()
 
-        if isinstance(initial_value, Dict):
-            for key, value in initial_value.items():
-                self[key] = value
-
         for key, value in kwargs.items():
             self[key] = value
 
-    def __str__(self):
-        return f'Struct {json.dumps(self.to_python(), indent=2, ensure_ascii=False)}'
+    def __contains__(self, item):
+        return item in self._fields
 
     def __setitem__(self, key: str, value: PyValue_T):
         self._fields[key] = Value.from_python(value)
@@ -49,7 +43,33 @@ class Struct(Message, MutableMapping):
     def __iter__(self) -> Iterator[PyValue_T]:
         return iter(self._fields)
 
-    def to_python(self) -> PyValue_T:
+    def __eq__(self, other: 'Struct') -> bool:
+        if not isinstance(other, Struct):
+            return False
+
+        return self._fields == other._fields
+
+    def _format(self, buffer: list = None, indent_level=0) -> str:
+        buffer = buffer or []
+
+        if indent_level == 0:
+            buffer.append(f"message {type(self).__name__}")
+
+        for key, value in self.items():
+            self._format_value(key, value, buffer, indent_level + 1)
+
+        return '\n'.join(buffer)
+
+    def set_value(self, key: str, value: 'Value'):
+        if not isinstance(value, Value):
+            raise ValueError(
+                "Expected a value of type 'Value', "
+                f"got value of type {type(value).__name__!r} instead"
+            )
+
+        self._fields[key] = value
+
+    def to_python(self) -> Dict[str, PyValue_T]:
         return {
             key: value.to_python()
             for key, value in self._fields.items()
@@ -57,7 +77,24 @@ class Struct(Message, MutableMapping):
 
     @classmethod
     def from_python(cls, value: dict) -> 'Struct':
-        return Struct(value)
+        return Struct(**value)
+
+    @classmethod
+    def from_stream(cls, stream: BinaryIO) -> 'Struct':
+        rv = cls()
+        map_field = cls._field_by_number[1]
+
+        while True:
+            # checking for end of message
+            try:
+                decode_header(stream)
+            except MessageDecodeError:
+                break
+
+            key, value = map_field.decode(stream)
+            rv.set_value(key, value)
+
+        return rv
 
 
 class NullValue(IntEnum):
@@ -82,7 +119,7 @@ class Value(Message):
     )
 
     @classmethod
-    def from_python(cls, value) -> 'Value':
+    def from_python(cls, value: PyValue_T) -> 'Value':
         if value is None:
             return Value(
                 null_value=NullValue.NULL_VALUE
